#!coding=utf-8
"""
This script is used to convert raw trained models to simplified and flexible
versions.
"""
from __future__ import print_function, absolute_import

import tensorflow as tf
import json
from . import pipeline
from . import kcnn
from os.path import join, dirname
from .constants import GHOST, VARIABLE_MOVING_AVERAGE_DECAY
from tensorflow.python.framework import graph_io
from tensorflow.python.tools import freeze_graph

__author__ = 'Xin Chen'
__email__ = 'Bismarrck@me.com'


FLAGS = tf.app.flags.FLAGS

tf.app.flags.DEFINE_string("checkpoint_dir", "./events",
                           """The directory where to read model checkpoints.""")
tf.app.flags.DEFINE_string('checkpoint', None,
                           """Specify the checkpoint to freeze.""")
tf.app.flags.DEFINE_string("aux_nodes", None,
                           """Comma separated string as the names of the 
                           auxiliary nodes to expose.""")

# TODO: test the batch normalization


def _get_transformer_repr(configs):
  """
  Return the JSON string for creating a `MultiTransformer`.

  Args:
    configs: a `dict` of configurations read from a JSON file.

  Returns:
    js: a `str` as the JSON string representation of the parameters for creating
      a `MultiTransformer`.

  """
  atom_types = configs["atom_types"]
  params = {"atom_types": [atom for atom in atom_types if atom != GHOST],
            "k_max": configs["k_max"],
            "norm": configs.get('norm', 'exp'),
            "norm_order": configs["norm_order"],
            "include_all_k": configs["include_all_k"],
            "periodic": configs["periodic"],
            "max_occurs": configs["max_occurs"],
            "atomic_forces": configs["atomic_forces_enabled"],
            "lj": configs.get("lj", False),
            "cutoff": configs.get("cutoff", None),
            "species": configs["species"]}
  return json.dumps(params)


def _get_output_node_names(forces=False):
  """
  Return the names of the tensors that should be accessed.
  """
  tensors = ["kCON/Energy/Sum/1_and_k", "kCON/Energy/y_contribs",
             "kCON/one-body/weights", "kCON/Energy/one-body/Conv2D",
             "transformer/json", "placeholders/inputs", "placeholders/occurs",
             "placeholders/weights", "placeholders/split_dims"]

  if forces:
    tensors.extend(["placeholders/coefficients", "placeholders/indexing",
                    "kCON/Forces/forces"])

  return ",".join(tensors)


def get_tensors_to_restore(forces=False):
  """
  Return a dict of (name, tensor_name) that should be restored from an exported
  model file.

  Returns:
    tensors: a `dict` as the tensors that should be restored.

  """
  return {"{}".format(name): "{}:0".format(name)
          for name in _get_output_node_names(forces=forces).split(",")}


def _inference(dataset_name, conv_sizes):
  """
  Inference a model of `KCNN` with inputs feeded from placeholders.

  Args:
    dataset_name: a `str` as the name of the dataset.
    conv_sizes: a `str` of comma-separated integers as the numbers of kernels.

  Returns:
    graph: a `tf.Graph` as the graph for inference.

  """
  graph = tf.Graph()

  with graph.as_default():

    configs = pipeline.get_configs(for_training=True, dataset_name=dataset_name)
    split_dims = configs["split_dims"]
    num_atom_types = configs["num_atom_types"]
    kbody_terms = [term.replace(",", "") for term in configs["kbody_terms"]]
    num_kernels = [int(units) for units in conv_sizes.split(",")]
    atomic_forces = configs["atomic_forces_enabled"]
    ck2 = configs["shape"][1]
    num_f, num_entries = configs["indexing_shape"]

    with tf.name_scope("transformer"):
      _ = tf.constant(_get_transformer_repr(configs), name="json")

    with tf.name_scope("placeholders"):
      inputs_ = tf.placeholder(
        tf.float32, shape=(None, 1, None, ck2), name="inputs")
      occurs_ = tf.placeholder(
        tf.float32, shape=(None, 1, 1, num_atom_types), name="occurs")
      weights_ = tf.placeholder(
        tf.float32, shape=(None, 1, None, 1), name="weights")
      split_dims_ = tf.placeholder(
        tf.int64, shape=(len(split_dims, )), name="split_dims")
      is_training_ = tf.placeholder(tf.bool, name="is_training")

      if atomic_forces:
        coef_ = tf.placeholder(
          tf.float32, shape=(None, None, ck2 * 6), name="coefficients")
        indexing_ = tf.placeholder(
          tf.int32, shape=(None, num_f, num_entries), name="indexing")
      else:
        coef_ = None
        indexing_ = None

    kcnn.kcnn(inputs_, occurs_, weights_, is_training=is_training_,
              split_dims=split_dims_, num_atom_types=num_atom_types,
              kbody_terms=kbody_terms, num_kernels=num_kernels, verbose=False,
              atomic_forces=atomic_forces, coefficients=coef_,
              indexing=indexing_, add_summary=False)

  return graph


def save_model(checkpoint_dir, dataset, conv_sizes, verbose=True,
               auxiliary_outputs=None):
  """
  take a GraphDef proto, a SaverDef proto, and a set of variable values stored
  in a checkpoint file, and output a GraphDef with all of the variable ops
  converted into const ops containing the values of the variables.

  Args:
    checkpoint_dir: a `str` as the directory to look for trained metadata.
    dataset: a `str` as the name of dataset to use.
    conv_sizes: a `str` of comma-separated integers as the numbers of kernels.
    verbose: a `bool` indicating whether or not should log the progress.
    auxiliary_outputs: a `List[str]` as the of additional tensors to expose.

  See Also:
    https://www.tensorflow.org/extend/tool_developers

  """
  graph = _inference(dataset, conv_sizes)

  with tf.Session(graph=graph,
                  config=tf.ConfigProto(allow_soft_placement=True)) as sess:

    # Restore the `moving averaged` model variables from the latest checkpoint.
    tf.global_variables_initializer().run()

    if FLAGS.checkpoint:
      checkpoint = FLAGS.checkpoint
      checkpoint_dir = dirname(checkpoint)
    else:
      ckpt = tf.train.get_checkpoint_state(checkpoint_dir)
      if ckpt and ckpt.model_checkpoint_path:
        checkpoint = ckpt.model_checkpoint_path
      else:
        raise IOError("Failed to restore the latest checkpoint!")

    variable_averages = tf.train.ExponentialMovingAverage(
      VARIABLE_MOVING_AVERAGE_DECAY)
    variables_to_restore = variable_averages.variables_to_restore()
    saver = tf.train.Saver(var_list=variables_to_restore, max_to_keep=1)
    saver.restore(sess, checkpoint)
    global_step = int(checkpoint.split('/')[-1].split('-')[-1])
    if verbose:
      print("Restore the checkpoint: {}".format(checkpoint))

    # Make the directory for saving exported models
    freeze_dir = join(checkpoint_dir, "freeze")
    if not tf.gfile.Exists(freeze_dir):
      tf.gfile.MakeDirs(freeze_dir)

    # Save the current session to a checkpoint so that values of variables can
    # be restored later.
    saver = tf.train.Saver(var_list=tf.global_variables())
    checkpoint_path = saver.save(
      sess,
      join(freeze_dir, "{}.ckpt".format(dataset)),
      global_step=global_step
    )

    # Write the graph to a GraphDef proto file
    graph_name = "{}-{}.pb".format(dataset, global_step)
    graph_path = join(freeze_dir, graph_name)
    graph_io.write_graph(graph, freeze_dir, name=graph_name)

    # Setup the configs and freeze the current graph
    output_node_names = _get_output_node_names(forces=FLAGS.forces)
    auxiliary_outputs = auxiliary_outputs or []
    if len(auxiliary_outputs) > 0:
      auxiliary_node_names = ",".join(auxiliary_outputs)
      output_node_names = ",".join([output_node_names, auxiliary_node_names])
    input_saver_def_path = ""
    restore_op_name = "save/restore_all"
    filename_tensor_name = "save/Const:0"
    clear_devices = False
    input_binary = False
    freeze_graph.freeze_graph(graph_path, input_saver_def_path,
                              input_binary, checkpoint_path,
                              output_node_names, restore_op_name,
                              filename_tensor_name, graph_path,
                              clear_devices, "")

    if verbose:
      print("Export the model to {}".format(graph_path))


# noinspection PyUnusedLocal,PyMissingOrEmptyDocstring
def main(unused):
  if FLAGS.aux_nodes is not None:
    aux_nodes = [name.strip() for name in FLAGS.aux_nodes.split(",")]
  else:
    aux_nodes = None
  save_model(FLAGS.checkpoint_dir, FLAGS.dataset, FLAGS.conv_sizes,
             auxiliary_outputs=aux_nodes)


if __name__ == "__main__":
  tf.app.run(main=main)



"""Use 'pip install .' to install."""


from os import path

from setuptools import find_packages, setup

if __name__ == '__main__':
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, 'docs', 'README.md'), encoding='utf-8') as f:
        long_description = f.read()

    tests_require = ['pytest-sugar', 'pytest-cov'],
    setup(name='kconmd',
          description='Molecular dynamics (MD) simulations supported by k-Bags Convolutional Neural Network (kcon).',
          keywords="molecular dynamics kcon",
          url='https://github.com/njzjz/kconmd',
          author='Jinzhe Zeng',
          author_email='jzzeng@stu.ecnu.edu.cn',
          packages=find_packages(),
          python_requires='~=3.6.0',
          install_requires=['numpy', 'scipy',
                            'matplotlib', 'scikit-learn', 'ase>=3.12',
                            'coloredlogs',
                            ],
          extras_require={
              "tf": ["tensorflow>=1.3"],
              "tf_gpu": ["tensorflow-gpu>=1.3"],
              'test': tests_require,
          },
          test_suite='kconmd.test',
          tests_require=tests_require,
          use_scm_version=True,
          setup_requires=['setuptools_scm', 'pytest-runner'],
          package_data={
              'kconmd': ['test/test.xyz', 'test/test.pb'],
          },
          long_description=long_description,
          long_description_content_type='text/markdown',
          classifiers=[
              "Natural Language :: English",
              "Operating System :: POSIX :: Linux",
              "Operating System :: Microsoft :: Windows",
              "Programming Language :: Python :: 3.6",
              "Topic :: Scientific/Engineering :: Chemistry",
              "Topic :: Software Development :: Libraries :: Python Modules",
              "Topic :: Software Development :: Version Control :: Git",
          ],
          zip_safe=True,
          )



# coding=utf-8
"""
This module is used for making predictions with trained models.
"""
from __future__ import print_function, absolute_import

import numpy as np
import tensorflow as tf
import json
from collections import Counter
from ase import Atoms
from ase.io.trajectory import Trajectory
from tensorflow.core.framework import graph_pb2
from tensorflow.python.framework import importer
from .constants import GHOST
from .save_model import get_tensors_to_restore
from .transformer import MultiTransformer, FixedLenMultiTransformer

__author__ = 'Xin Chen'
__email__ = 'Bismarrck@me.com'


def restore_transformer(graph, session, fixed=False):
  """
  Restore a `MultiTransformer` from the freezed graph.

  Args:
    graph: a `tf.Graph`.
    session: a `tf.Session` to execute ops.
    fixed: a `bool`. If True, a `FixedLenMultiTransformer` will be restored.
      Otherwise a `MultiTransformer` will be restored.

  Returns:
    clf: a `MultiTransformer` or a `FixedLenMultiTransformer`.

  """
  tensor = graph.get_tensor_by_name("transformer/json:0")
  params = dict(json.loads(session.run(tensor).decode()))
  if not fixed:
    return MultiTransformer(**{k: v for k, v in params.items()
                               if k != "species"})
  else:
    max_occurs = Counter(params["species"])
    kwargs = {k: v for k, v in params.items()
              if k not in ("species", "atom_types", "max_occurs")}
    return FixedLenMultiTransformer(max_occurs, **kwargs)


class KcnnPredictor:
  """
  An energy predictor based on the deep neural network of 'KCNN'.
  """

  def __init__(self, graph_model_path, fixed=False):
    """
    Initialization method.

    Args:
      graph_model_path: a `str` as the freezed graph model to load.
      fixed: a `bool`. If True, a `FixedLenMultiTransformer` will be restored.
        Otherwise a `MultiTransformer` will be restored.

    """

    graph = tf.Graph()

    with graph.as_default():
      output_graph_def = graph_pb2.GraphDef()
      with open(graph_model_path, "rb") as f:
        output_graph_def.ParseFromString(f.read())
        importer.import_graph_def(output_graph_def, name="")

    self._graph = graph
    self._sess = tf.Session(graph=graph,config=tf.ConfigProto(intra_op_parallelism_threads=2,inter_op_parallelism_threads=5))
    self._transformer = restore_transformer(self._graph, self._sess, fixed)
    assert isinstance(self._transformer, MultiTransformer)

    self._initialize_tensors()
    self._y_atomic_1body = self._get_y_atomic_1body(
      self._transformer.atom_types
    )

  @property
  def k_max(self):
    """
    Return the many-body expansion factor for this model.
    """
    return self._transformer.k_max

  @property
  def included_k(self):
    """
    Return the included k under the many body expansion scheme.
    """
    return self._transformer.included_k

  @property
  def is_periodic(self):
    """
    Return True if this model is used for periodic structures.
    """
    return self._transformer.is_periodic

  @property
  def sess(self):
    """
    Return the associated session for this predictor.
    """
    return self._sess

  @property
  def graph(self):
    """
    Return the restored freezed graph.
    """
    return self._graph

  @property
  def supported_stoichiometries(self):
    """
    Return the supported stoichiometries.
    """
    supported = {}
    max_occurs = self._transformer.max_occurs
    for atom, max_occur in max_occurs.items():
      if atom != GHOST:
        supported[atom] = [0, max_occur]
    return supported

  @property
  def kbody_terms(self):
    """
    Return the associated kbody terms.
    """
    return self._transformer.kbody_terms

  @property
  def atomic_forces(self):
    """
    Return True if this predictor supports predicting atomic forces.

    Notes:
      Currently the atomic forces are not supported!

    """
    return self._transformer.atomic_forces_enabled

  @property
  def transformer(self):
    """
    Return the `MultiTransformer` of this predictor.
    """
    return self._transformer

  def __str__(self):
    """
    Return the string representation of this predictor.
    """
    species = list(self._transformer.species)
    if GHOST in species:
      species.remove(GHOST)
    if not self.is_periodic:
      outputs = ["KcnnPredictor of {}".format("".join(species))]
    else:
      outputs = ["Periodic KcnnPredictor of {}".format("".join(species))]
    for atom, (nmin, nmax) in self.supported_stoichiometries.items():
      outputs.append("  {:2s} : [{}, {}]".format(atom, nmin, nmax))
    outputs.append("End")
    return "\n".join(outputs)

  def _initialize_tensors(self):
    """
    Initialize the tensors.
    """
    tensors = {}
    forces = self.atomic_forces
    for name, tensor_name in get_tensors_to_restore(forces=forces).items():
      tensors[name] = self._graph.get_tensor_by_name(tensor_name)

    # The operators
    self._operator_y_nn = tensors["kCON/Energy/Sum/1_and_k"]
    self._operator_y_kbody = tensors["kCON/Energy/y_contribs"]
    self._operator_y_1body = tensors["kCON/Energy/one-body/Conv2D"]

    # The data tensor
    self._tensor_1body = tensors["kCON/one-body/weights"]

    # Placeholders
    self._placeholder_inputs = tensors["placeholders/inputs"]
    self._placeholder_occurs = tensors["placeholders/occurs"]
    self._placeholder_weights = tensors["placeholders/weights"]
    self._placeholder_split_dims = tensors["placeholders/split_dims"]

    # Tensors for predicting atomic forces
    if self._transformer.atomic_forces_enabled:
      self._placeholder_coefficients = tensors["placeholders/coefficients"]
      self._placeholder_indexing = tensors["placeholders/indexing"]
      self._operator_f_nn = tensors["kCON/Forces/forces"]
    else:
      self._placeholder_coefficients = None
      self._placeholder_indexing = None
      self._operator_f_nn = None

    # Temporarily disable this tensor. I think this will be removed soon.
    # self._placeholder_is_training = tensors["placeholders/is_training"]

  def _get_y_atomic_1body(self, species):
    """
    Return the one-body energy for each type of atom.
    
    Args:
      species: a `List[str]` as the ordered species for this model.
    
    Returns:
      y_atomic_1body: a `Dict[str, float]` as the 1body energy for 
        each kind of atom.
    
    """
    weights = self._sess.run(self._tensor_1body)
    return dict(zip(species, weights.flatten().tolist()))

  def get_feed_dict(self, atoms_or_trajectory):
    """
    Return the feed dict for the inputs.

    Args:
      atoms_or_trajectory: an `ase.Atoms` or an `ase.io.TrajectoryReader` or a
        list of `ase.Atoms` with the same stoichiometry.

    Returns:
      species: a list of `str` as the stoichiometry.
      feed_dict: a `dict` as the feed dict to run.

    """
    assert isinstance(self._transformer, MultiTransformer)

    if isinstance(atoms_or_trajectory, Atoms):
      transform_func = self._transformer.transform
      ntotal = 1
      species = atoms_or_trajectory.get_chemical_symbols()
    elif isinstance(atoms_or_trajectory, (list, tuple, Trajectory)):
      transform_func = self._transformer.transform_trajectory
      ntotal = len(atoms_or_trajectory)
      species = atoms_or_trajectory[0].get_chemical_symbols()
    else:
      raise ValueError("`atoms_or_trajectory` should be an `ase.Atoms` or an"
                       "`ase.io.TrajectoryReader` or a list of `ase.Atoms`!")

    # The `transformed` is a named tupe: KcnnSample.
    transformed = transform_func(atoms_or_trajectory)

    # Build the feed dict for running the session.
    ck2 = self._transformer.ck2
    features = transformed.features.reshape((ntotal, 1, -1, ck2))
    weights = transformed.binary_weights.reshape((ntotal, 1, -1, 1))
    occurs = transformed.occurs.reshape((ntotal, 1, 1, -1))
    split_dims = transformed.split_dims

    if self._transformer.atomic_forces_enabled:
        coefficients = transformed.coefficients.reshape((ntotal, -1, ck2 * 6))
        indexing = transformed.indexing.reshape((ntotal, 3*(len(self._transformer._species)-self._transformer._num_ghosts), -1))
        feed_dict = {self._placeholder_inputs: features,
                     self._placeholder_occurs: occurs,
                     self._placeholder_weights: weights,
                     self._placeholder_split_dims: split_dims,
                     self._placeholder_coefficients: coefficients,
                     self._placeholder_indexing: indexing}
    else:
        feed_dict = {self._placeholder_inputs: features,
                     self._placeholder_occurs: occurs,
                     self._placeholder_weights: weights,
                     self._placeholder_split_dims: split_dims}
    return species, feed_dict

  def predict_total_energy(self, atoms_or_trajectory):
    """
    Only make predictions of total energies. All input structures must have the
    same kind of atomic species.

    Args:
      atoms_or_trajectory: an `ase.Atoms` or an `ase.io.TrajectoryReader` or a
        list of `ase.Atoms` with the same stoichiometry.

    Returns:
      y_total: a 1D array of shape `[num_examples, ]` as the total energies.

    """
    _, feed_dict = self.get_feed_dict(atoms_or_trajectory)
    y_total = self._sess.run(self._operator_y_nn, feed_dict=feed_dict)
    return np.negative(y_total)

  def predict(self, atoms_or_trajectory):
    """
    Make the prediction for the given structures. All input structures must have
    the same kind of atomic species.

    Args:
      atoms_or_trajectory: an `ase.Atoms` or an `ase.io.TrajectoryReader` or a
        list of `ase.Atoms` with the same stoichiometry.

    Returns:
      y_total: a `float32` array of shape `[num_examples, ]` as the predicted
        total energies.
      y_1body: a `float32` array of shape `[num_examples, ]` as the predicted
        one-body energies.
      y_atomics: a `float32` array of shape `[num_examples, num_atoms]` as the
        predicted energies for atoms which sum up to total energies.
      y_kbody: a `float32` array of shape `[num_examples, D]` as the kbody
        contribs. `D` is the total dimension.

    """
    species, feed_dict = self.get_feed_dict(atoms_or_trajectory)

    # Run the operations to get the predicted energies.
    y_total, y_kbody, y_1body = self._sess.run(
      [self._operator_y_nn, self._operator_y_kbody, self._operator_y_1body],
      feed_dict=feed_dict
    )
    y_1body = np.squeeze(y_1body)

    # Compute the atomic energies from kbody contribs.
    y_kbody = np.squeeze(y_kbody, axis=(1, 3))
    y_atomic = self._transformer.compute_atomic_energies(
      species, y_kbody, self._y_atomic_1body)

    return (np.negative(np.atleast_1d(y_total)),
            np.negative(np.atleast_1d(y_1body)),
            np.negative(y_atomic),
            np.negative(y_kbody))

  def predict_forces(self, atoms_or_trajectory):
    """
    Only make predictions of atomic forces. All input structures must have the
    same kind of atomic species.

    Args:
      atoms_or_trajectory: an `ase.Atoms` or an `ase.io.TrajectoryReader` or a
        list of `ase.Atoms` with the same stoichiometry.

    Returns:
      f_atomics: a `float32` array of shape `[num_examples, num_atoms * 3]`
        as the predicted atomic forces.

    """
    # predict forces for trajectory is broken now
    _, feed_dict = self.get_feed_dict(atoms_or_trajectory)
    f_atomics = self._sess.run(self._operator_f_nn, feed_dict=feed_dict)
    return f_atomics

  def eval(self, name, feed_dict=None):
    """
    Evaluate a specific tensor from the graph given the feed dict.

    Args:
      name: a `str` as the name of the tensor.
      feed_dict: a `dict` as the feed dict.

    Returns:
      results: a `object` as the evaluation result.

    """
    try:
      tensor = self._graph.get_tensor_by_name(name)
    except Exception:
      raise KeyError(
        "The tensor {} can not be found in the graph!".format(name))
    else:
      return self.sess.run(tensor, feed_dict=feed_dict)



import numpy as np
from ase import Atoms


class Feed:
    def __init__(self, transformer, cutoff):
        self.cutoff = cutoff
        self._transformer = transformer
        self.maxatoms = self._transformer.max_occurs

    def runmp(self, atoms, pool):
        self.atoms = atoms
        self.N = len(atoms)
        feeds = pool.imap(self.getfeedfromatoms, np.arange(self.N))
        return feeds

    def getfeedfromatoms(self, i):
        traj, index = self.gettraj(i)
        feed = self.get_feed_dict(traj, index)
        return feed

    def get_feed_dict(self, atoms_or_trajectory, index):
        transform_func = self._transformer.transform
        species = atoms_or_trajectory.get_chemical_symbols()
        transformed = transform_func(atoms_or_trajectory)
        ck2 = self._transformer.ck2
        features = transformed.features.reshape((1, 1, -1, ck2))
        weights = transformed.binary_weights.reshape((1, 1, -1, 1))
        occurs = transformed.occurs.reshape((1, 1, 1, -1))
        split_dims = transformed.split_dims
        coefficients = transformed.coefficients.reshape((1, -1, ck2 * 6))
        indexing = transformed.indexing.reshape(
            (1, 3*(len(self._transformer._species)-self._transformer._num_ghosts), -1))
        feed_dict = {"inputs": features,
                     "occurs": occurs,
                     "weights": weights,
                     "split_dims": split_dims,
                     "coefficients": coefficients,
                     "indexing": indexing,
                     "index": index}
        return feed_dict

    def gettraj(self, i):
        distances = self.atoms.get_distances(i, np.arange(self.N), mic=True)
        distances_cutoff = [(distances[i], i) for i in np.arange(
            self.N) if distances[i] < self.cutoff] if self.cutoff > 0 else distances
        distances_cutoff.sort()
        atom_num = {}
        indexs = []
        for distance in distances_cutoff:
            symbol = self.atoms[distance[1]].symbol
            if symbol in atom_num:
                if atom_num[symbol] < self.maxatoms[symbol]:
                    atom_num[symbol] += 1
                    indexs.append(distance[1])
            else:
                if symbol in self.maxatoms:
                    atom_num[symbol] = 1
                    indexs.append(distance[1])
        traj = Atoms([self.atoms[j]
                      for j in indexs], cell=self.atoms.get_cell())
        traj.wrap(center=self.atoms[i].position/self.atoms.get_cell_lengths_and_angles()[
                  0:3], pbc=self.atoms.get_pbc())
        index = indexs.index(i)
        return traj, index



from ..force import ComputeForces
import time
import numpy as np
import logging


class kconMD:
    def __init__(self, pbfilename, xyzfilename="comb.xyz",
                 outputfilename="force.dat", cell=None, pbc=False, cutoff=6,
                 unit=1, vdw=False, nproc=None):
        self.pbfilename = pbfilename
        self.xyzfilename = xyzfilename
        self.cell = cell if cell else [0, 0, 0]
        self.pbc = pbc
        self.cutoff = cutoff
        self.outputfilename = outputfilename
        self.unit = unit
        self.vdw = vdw
        self.cf = None
        self.nproc = nproc

    def initcf(self):
        if self.cf == None:
            self.cf = ComputeForces(
                self.pbfilename, cell=self.cell, pbc=self.pbc,
                cutoff=self.cutoff, vdw=self.vdw, nproc=self.nproc)

    def printforce(self):
        time1 = time.time()
        self.initcf()
        forces = self.cf.predictforcesfromxyz(self.xyzfilename)
        forces *= self.unit
        np.savetxt(self.outputfilename, forces, fmt='%16.9f')
        time2 = time.time()
        logging.info(f"Compute Forces: Time cosumed: {time2-time1:.3f} s")



# coding=utf-8
"""
An implementation of the AMSGrad optimizer for tensorflow.

https://github.com/taki0112/AMSGrad-Tensorflow

The paper: On the Convergence of Adam and Beyond
"""
from tensorflow.python.eager import context
from tensorflow.python.framework import ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import resource_variable_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.ops import variable_scope
from tensorflow.python.training import optimizer


class AmsGrad(optimizer.Optimizer):
  """
  The AMSGrad optimizer.
  """

  def __init__(self, learning_rate=0.01, beta1=0.9, beta2=0.99, epsilon=1e-8,
               use_locking=False, name="amsgrad"):
    """
    Initialization method.
    """
    super(AmsGrad, self).__init__(use_locking, name)
    self._lr = learning_rate
    self._beta1 = beta1
    self._beta2 = beta2
    self._epsilon = epsilon
    self._lr_t = None
    self._beta1_t = None
    self._beta2_t = None
    self._epsilon_t = None
    self._beta1_power = None
    self._beta2_power = None

  def _create_slots(self, var_list):
    first_var = min(var_list, key=lambda x: x.name)

    create_new = self._beta1_power is None
    if not create_new and not context.in_eager_mode():
      create_new = (self._beta1_power.graph is not first_var.graph)

    if create_new:
      with ops.colocate_with(first_var):
        self._beta1_power = variable_scope.variable(
          self._beta1, name="beta1_power", trainable=False)
        self._beta2_power = variable_scope.variable(
          self._beta2, name="beta2_power", trainable=False)

    # Create slots for the first and second moments.
    for v in var_list:
      self._zeros_slot(v, "m", self._name)
      self._zeros_slot(v, "v", self._name)
      self._zeros_slot(v, "vhat", self._name)

  def _prepare(self):
    self._lr_t = ops.convert_to_tensor(self._lr)
    self._beta1_t = ops.convert_to_tensor(self._beta1)
    self._beta2_t = ops.convert_to_tensor(self._beta2)
    self._epsilon_t = ops.convert_to_tensor(self._epsilon)

  def _apply_dense(self, grad, var):
    beta1_power = math_ops.cast(self._beta1_power, var.dtype.base_dtype)
    beta2_power = math_ops.cast(self._beta2_power, var.dtype.base_dtype)
    lr_t = math_ops.cast(self._lr_t, var.dtype.base_dtype)
    beta1_t = math_ops.cast(self._beta1_t, var.dtype.base_dtype)
    beta2_t = math_ops.cast(self._beta2_t, var.dtype.base_dtype)
    epsilon_t = math_ops.cast(self._epsilon_t, var.dtype.base_dtype)

    lr = (lr_t * math_ops.sqrt(1 - beta2_power) / (1 - beta1_power))

    # m_t = beta1 * m + (1 - beta1) * g_t
    m = self.get_slot(var, "m")
    m_scaled_g_values = grad * (1 - beta1_t)
    m_t = state_ops.assign(m, beta1_t * m + m_scaled_g_values,
                           use_locking=self._use_locking)

    # v_t = beta2 * v + (1 - beta2) * (g_t * g_t)
    v = self.get_slot(var, "v")
    v_scaled_g_values = (grad * grad) * (1 - beta2_t)
    v_t = state_ops.assign(v, beta2_t * v + v_scaled_g_values,
                           use_locking=self._use_locking)

    # amsgrad
    vhat = self.get_slot(var, "vhat")
    vhat_t = state_ops.assign(vhat, math_ops.maximum(v_t, vhat))
    v_sqrt = math_ops.sqrt(vhat_t)

    var_update = state_ops.assign_sub(var, lr * m_t / (v_sqrt + epsilon_t),
                                      use_locking=self._use_locking)
    return control_flow_ops.group(*[var_update, m_t, v_t, vhat_t])

  def _resource_apply_dense(self, grad, var):
    var = var.handle
    beta1_power = math_ops.cast(self._beta1_power, grad.dtype.base_dtype)
    beta2_power = math_ops.cast(self._beta2_power, grad.dtype.base_dtype)
    lr_t = math_ops.cast(self._lr_t, grad.dtype.base_dtype)
    beta1_t = math_ops.cast(self._beta1_t, grad.dtype.base_dtype)
    beta2_t = math_ops.cast(self._beta2_t, grad.dtype.base_dtype)
    epsilon_t = math_ops.cast(self._epsilon_t, grad.dtype.base_dtype)

    lr = (lr_t * math_ops.sqrt(1 - beta2_power) / (1 - beta1_power))

    # m_t = beta1 * m + (1 - beta1) * g_t
    m = self.get_slot(var, "m").handle
    m_scaled_g_values = grad * (1 - beta1_t)
    m_t = state_ops.assign(m, beta1_t * m + m_scaled_g_values,
                           use_locking=self._use_locking)

    # v_t = beta2 * v + (1 - beta2) * (g_t * g_t)
    v = self.get_slot(var, "v").handle
    v_scaled_g_values = (grad * grad) * (1 - beta2_t)
    v_t = state_ops.assign(v, beta2_t * v + v_scaled_g_values,
                           use_locking=self._use_locking)

    # amsgrad
    vhat = self.get_slot(var, "vhat").handle
    vhat_t = state_ops.assign(vhat, math_ops.maximum(v_t, vhat))
    v_sqrt = math_ops.sqrt(vhat_t)

    var_update = state_ops.assign_sub(var, lr * m_t / (v_sqrt + epsilon_t),
                                      use_locking=self._use_locking)
    return control_flow_ops.group(*[var_update, m_t, v_t, vhat_t])

  def _apply_sparse_shared(self, grad, var, indices, scatter_add):
    beta1_power = math_ops.cast(self._beta1_power, var.dtype.base_dtype)
    beta2_power = math_ops.cast(self._beta2_power, var.dtype.base_dtype)
    lr_t = math_ops.cast(self._lr_t, var.dtype.base_dtype)
    beta1_t = math_ops.cast(self._beta1_t, var.dtype.base_dtype)
    beta2_t = math_ops.cast(self._beta2_t, var.dtype.base_dtype)
    epsilon_t = math_ops.cast(self._epsilon_t, var.dtype.base_dtype)

    lr = (lr_t * math_ops.sqrt(1 - beta2_power) / (1 - beta1_power))

    # m_t = beta1 * m + (1 - beta1) * g_t
    m = self.get_slot(var, "m")
    m_scaled_g_values = grad * (1 - beta1_t)
    m_t = state_ops.assign(m, m * beta1_t, use_locking=self._use_locking)
    with ops.control_dependencies([m_t]):
      m_t = scatter_add(m, indices, m_scaled_g_values)

    # v_t = beta2 * v + (1 - beta2) * (g_t * g_t)
    v = self.get_slot(var, "v")
    v_scaled_g_values = (grad * grad) * (1 - beta2_t)
    v_t = state_ops.assign(v, v * beta2_t, use_locking=self._use_locking)
    with ops.control_dependencies([v_t]):
      v_t = scatter_add(v, indices, v_scaled_g_values)

    # amsgrad
    vhat = self.get_slot(var, "vhat")
    vhat_t = state_ops.assign(vhat, math_ops.maximum(v_t, vhat))
    v_sqrt = math_ops.sqrt(vhat_t)
    var_update = state_ops.assign_sub(var, lr * m_t / (v_sqrt + epsilon_t),
                                      use_locking=self._use_locking)
    return control_flow_ops.group(*[var_update, m_t, v_t, vhat_t])

  def _apply_sparse(self, grad, var):
    return self._apply_sparse_shared(
      grad.values,
      var,
      grad.indices,
      lambda x, i, v: state_ops.scatter_add(x, i, v,
                                            use_locking=self._use_locking)
    )

  @staticmethod
  def _resource_scatter_add(x, i, v):
    dependencies = [resource_variable_ops.resource_scatter_add(x.handle, i, v)]
    with ops.control_dependencies(dependencies):
      return x.value()

  def _resource_apply_sparse(self, grad, var, indices):
    return self._apply_sparse_shared(
      grad, var, indices, self._resource_scatter_add)

  def _finish(self, update_ops, name_scope):
    # Update the power accumulators.
    with ops.control_dependencies(update_ops):
      with ops.colocate_with(self._beta1_power):
        update_beta1 = self._beta1_power.assign(
          self._beta1_power * self._beta1_t, use_locking=self._use_locking)
        update_beta2 = self._beta2_power.assign(
          self._beta2_power * self._beta2_t, use_locking=self._use_locking)
      return control_flow_ops.group(
        *update_ops + [update_beta1, update_beta2],
        name=name_scope
      )
